/*

    sftpd - meant to be a small and flexible ftp server implementation
    Copyright (C) 2009  Jan Murawski  der@oelgoetze.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

    For more information about the GNU Public license see:
    http://www.gnu.org/copyleft/gpl.html

*/

#include "directoryObject.h"

//
// init directory object with given values
//
directoryObj::directoryObj(string directory, string writable)
{
    if (writable == "rw")
        this->writable = true;
    else
        this->writable = false;

    // init our root directory entry
    this->rootNode = new dirNode;
    this->rootNode->rootDir = directory;
    this->rootNode->virtuaRootDir = "/";
    this->rootNode->previousNode = 0;
    this->rootNode->nextNode = 0;

    currVirtualDir = "/";
    currDir = directory;
}

directoryObj::~directoryObj()
{
    // clean up all dir nodes used to trace current path
    removeSubnodes();
    delete this->rootNode;
}


//
// removes all dir nodes except for head node
//
void directoryObj::removeSubnodes()
{
    dirNode *node = this->rootNode;
    int subNodes = 0;
    
    while (node->nextNode != 0)
    {
        node = node->nextNode;
        subNodes++;
    }

    for (int i = 0; i < subNodes; i++)
    {
        node = node->previousNode;
        delete node ->nextNode;
    }
}


//
// steps a dir up. removes last dir-node, if no more needed
// returns false, whend absolut root is reached
//
bool directoryObj::goDirUp()
{
    dirNode *lastNode = this->rootNode;
    while (lastNode->nextNode != 0)
        lastNode = lastNode->nextNode;

    // do we need to get a node upwards?
    if (lastNode->virtuaRootDir == this->currVirtualDir)
    {
        if (lastNode == this->rootNode)
        {
            // we are at absolut root. nothing left to do
            return false;
        }
        else
        {
            lastNode->previousNode->nextNode = 0;
            delete lastNode;
        }
    }
    removeLastDirFromPath(this->currVirtualDir, true);
    removeLastDirFromPath(this->currDir, false);
    return true;
}


//
// change to next dir
//
bool directoryObj::goDirDown(string folderName)
{
    if (folderName == "../")
        return this->goDirUp();

    // ensure incomming path does not contain any slashes
    unsigned int pos;
    while ((pos = folderName.find("/")) != string::npos)
        folderName.erase(pos, 1);

    finalSlashToPath();
    finalBackSlashToPath();

    // FIXME: support for dirNodes and .lnk files must be added here

    // change to dir if it exists on disk
    if (fileExists(this->currDir + folderName))
    {
        this->currDir += folderName;
        this->currVirtualDir += folderName;

        return true;
    }
    // changing folder failed
    else
        return false;
}


//
// if any more existant, remove last folder from path
// designed for absout paths only
// returns true, when dir removed
//
bool directoryObj::removeLastDirFromPath(string &path, bool unixStyle)
{
    int pos;

    if (unixStyle)
    {
        if (path.length() == 1)  //  means = "/"
            return false;

        pos = path.find_last_of("/");
        if (pos +1 == path.length())
        {
            path.erase(path.length() -1);
            pos = path.find_last_of("/");
        }
        path.erase(pos +1);
    }
    else
    {
        if (path.length() == 3)  // means = "x:\"
            return false;

        pos = path.find_last_of("\\");
        if (pos +1 == path.length())
        {
            path.erase(path.length() -1);
            pos = path.find_last_of("\\");
        }
        path.erase(pos);
    }

    return true;
}


//
// returns pwd
// pwd is changend upwards until a folder is reached
// which is existant
//
bool directoryObj::getPwd(string &pwd)
{
    if (fileExists(this->currDir))
    {
        // ensure virtual path has final slash
        finalSlashToPath();

        pwd = this->currVirtualDir;
        return true;
    }
    else
    {
        bool headDirsLeft;
        do
        {
            headDirsLeft = goDirUp();
            pwd = this->currVirtualDir;
        }
        while (fileExists(this->currDir) && headDirsLeft);

        return false;
    }
}


//
// returns a complete directory listing
//
bool directoryObj::getDirListing(string userName, vector<string*> *lines)
{
    // if dir is no more existant, return
    if (!fileExists(this->currDir))
        return false;

    // append final backslash to path if non existant
    finalBackSlashToPath();
    
    // copy dirname to buffer
    char *filename = new char[this->currDir.size() +1 +1];
    for (unsigned int i = 0; i < this->currDir.size(); i++)
        *(filename + i) = this->currDir.at(i);
    *(filename + this->currDir.size()) = '*';
    *(filename + this->currDir.size() +1) = 0;
    
    // list, to store all files sorted
    deque<searchEntry*> entrys;

    // remeber longest file length for later formating
    int fileSizeMaxLength = 0;

    // get current time
    SYSTEMTIME currTime;
    GetLocalTime(&currTime);

    // enumerate all files an dirs an save to dequ
    {
        // vars for searching
        WIN32_FIND_DATA FindFileData;
        HANDLE hFind = INVALID_HANDLE_VALUE;
        DWORD dwError;

        // Find the first file in the directory.
        hFind = FindFirstFile(filename, &FindFileData);

        if (hFind == INVALID_HANDLE_VALUE) 
        {
            //printf ("Invalid file handle. Error is %u.\n", GetLastError());
            // FIXME: error logging
            goto returnFalse;
        } 
        else 
        {
            // List all the other files in the directory.
            do 
            {
                if (strcmp(FindFileData.cFileName, ".") == 0  || strcmp(FindFileData.cFileName, "..") == 0)
                    continue;

                // create new entry for sorting
                searchEntry *entry = new searchEntry;
    
                // filename
                unsigned int len = strlen(FindFileData.cFileName);
                for (unsigned int i = 0; i < len; i++)
                    entry->fileName += *(FindFileData.cFileName + i);

                {
                    SYSTEMTIME systime, systimeUTC; 
                    // convert filetime to UTC
                    // and adjust acording to timezone
                    if (!FileTimeToSystemTime(&FindFileData.ftLastWriteTime, &systimeUTC)
                        || !SystemTimeToTzSpecificLocalTime(0, &systimeUTC, &systime))
                    {
                        // FIXME: error logging
                        goto returnFalse;
                    }

                    switch(systime.wMonth)
                    {
                        case 1:
                            entry->timeStamp = "Jan";
                            break;
                        case 2:
                            entry->timeStamp = "Feb";
                            break;
                        case 3:
                            entry->timeStamp = "Mar";
                            break;
                        case 4:
                            entry->timeStamp = "Apr";
                            break;
                        case 5:
                            entry->timeStamp = "May";
                            break;
                        case 6:
                            entry->timeStamp = "Jun";
                            break;
                        case 7:
                            entry->timeStamp = "Jul";
                            break;
                        case 8:
                            entry->timeStamp = "Aug";
                            break;
                        case 9:
                            entry->timeStamp = "Sep";
                            break;
                        case 10:
                            entry->timeStamp = "Oct";
                            break;
                        case 11:
                            entry->timeStamp = "Nov";
                            break;
                        case 12:
                            entry->timeStamp = "Dec";
                            break;
                    }

                    entry->timeStamp += ' ';
                    if (intToStr(systime.wDay).length() == 1)
                        entry->timeStamp += ' ';
                    entry->timeStamp += intToStr(systime.wDay);
                    
                    // append year, if time stamp is older than 6 month
                    // else append time of change
                    {
                        entry->timeStamp += ' ';

                        // reduce current time by 6 months
                        if (currTime.wMonth < 6)
                        {
                            currTime.wMonth = 12 - (6 - currTime.wMonth);
                            currTime.wYear -= 1;
                        }
                        else
                            currTime.wMonth -= 6;

                        if (compareTimes(&systime, &currTime) == -1)
                        { // is older than 6 months
                            entry->timeStamp += ' ';
                            entry->timeStamp += intToStr(systime.wYear);

                        }
                        else // is newer than 6 months
                        {
                            string hour = intToStr(systime.wHour);
                            string minute = intToStr(systime.wMinute);

                            if (hour.size() < 2)
                                hour = '0' + hour;
                            if (minute.size() < 2)
                                minute = '0' + minute;

                            entry->timeStamp += hour + ':' + minute;
                        }
                    }
                }

                // is directpory?
                if ((FindFileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == FILE_ATTRIBUTE_DIRECTORY)
                    entry->isDirectory = true;
                else
                    entry->isDirectory = false;

                // is writable?
                if ((FindFileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == FILE_ATTRIBUTE_READONLY)
                    entry->isReadOnly = true;
                else
                    entry->isReadOnly = false;

                // is executable?
                if (IsExecutable(this->currDir + FindFileData.cFileName))
                    entry->isExecutable = true;
                else
                    entry->isExecutable = false;

                // filesize
                if (entry->isDirectory)
                    entry->fileSize = "0";
                else
                {
                    if (FindFileData.nFileSizeHigh == 0)
                        entry->fileSize = intToStr((int) FindFileData.nFileSizeLow);
                    else
                        entry->fileSize = intToStr(
                            (long long) FindFileData.nFileSizeHigh * 
                            ((long long) MAXDWORD + 1) + 
                            FindFileData.nFileSizeLow);
                }

                // remember literally longest filesize for formating
                int tmp_fileSizeMaxLength;
                if ((tmp_fileSizeMaxLength = entry->fileSize.size()) > fileSizeMaxLength)
                    fileSizeMaxLength = tmp_fileSizeMaxLength;

                // generate the access rights for file / folder
                {
                    string user, group, others;

                    // file should always be readeble
                    user = 'r';     
                    // file readonly or only reading allowed for user?
                    if (entry->isReadOnly || !this->writable)
                        user += '-';
                    else
                        user += 'w';
                    // file executable or directory?
                    if (entry->isExecutable || entry->isDirectory)
                        user += 'x';
                    else
                        user += '-';

                    // no groups implemented yet
                    group = others = "---";

                    // directory flag
                    if (entry->isDirectory)
                        entry->dirMask = 'd';
                    else
                        entry->dirMask = '-';

                    // assemble access rights
                    entry->dirMask += user + group + others;
                }

                // create lower case duplicate of string for the comparisson
                entry->lowerCaseFileName = "";
                for (unsigned int i = 0; i < entry->fileName.size(); i++)
                    entry->lowerCaseFileName += tolower(entry->fileName[i]);

                // search for correct position in listing
                unsigned int pos = 0;
                for (unsigned int i = 0; i < entrys.size(); i++)
                {
                    if (entry->isDirectory)                 // we "are" directory
                    {
                        if (!entrys.at(i)->isDirectory)     // current element is not directory
                        {
                            pos = i;
                            break;                          // => insert us here
                        }
                    }
                    else                                    // we "are" not a directory
                        if (entrys.at(i)->isDirectory)      // current element is
                        {
                            pos = i;
                            continue;                       // => skip this (=> all) directory entrys
                        }

                    // if our string is smaller than the current one, than this is our positon
                    if (entry->lowerCaseFileName.compare(entrys.at(i)->lowerCaseFileName) < 0)
                        break;

                    pos = i;
                }
                // insert at determined position
                entrys.insert(entrys.begin() + pos, entry);     
            }
            while (FindNextFile(hFind, &FindFileData) != 0);    // another file found?

            dwError = GetLastError();
            FindClose(hFind);   // search is finished
            if (dwError != ERROR_NO_MORE_FILES) 
            {
                //printf ("FindNextFile error. Error is %u.\n", dwError);
                // FIXME: error logging
                goto returnFalse;
            }
        }
    }

    lines->resize(entrys.size());
    for (unsigned int i = 0; i < entrys.size(); i++)
    {
        // get curr entry to local var
        searchEntry *entry = entrys.at(i);

        string *str = new string;

        // file file size with blanks
        fillString(&entry->fileSize, ' ', fileSizeMaxLength);

        // assemble unix style ls -l line for file
        *str = entry->dirMask + " 1 " + userName + " ftp " + entry->fileSize;
        *str += ' ' + entry->timeStamp + ' ' + entry->fileName + "\r\n";

        // and place it in vector
        lines->at(i) = str;

        // clean up entry, used for information collecting
        delete entry;
    }

    delete filename;
    return true;

returnFalse:

    for (unsigned int i = 0; i < entrys.size(); i++)
        delete entrys.at(i);

    delete filename;
    return false;
}


//
// changes to a given relative or absolut path
//
bool directoryObj::changeDir(string vDir)
{
    if (vDir.length() == 0)
        return true;

    // ensure ur current virtual folder 
    // and given paht end with /
    finalSlashToPath();
    finalSlashToPath(vDir);

    // if path is absolute strip folders, we are already in
    if (vDir.substr(0, 1) == "/")
    {
        string currDir = this->currVirtualDir;      // tmp copy of current dir
        deque<string> dirsToGoUp;           // remember all dirs to leave

        // remove dirs, until current dir is part of target dir
        while (vDir.find(currDir) == string::npos)
        {
            // remove trailing slash if existant
            if (currDir.find_last_of("/") == currDir.length() -1)
                currDir.erase(currDir.length() -1);
            
            // locate beginning of last dir name
            unsigned int pos = currDir.find_last_of("/");
            // save last dir name
            dirsToGoUp.push_front(currDir.substr(pos +1));
            // and remove it from tmp
            currDir.erase(pos);
        }

        // switch needed number of folders up
        // (and hope function does not fail)
        for (unsigned int i = 0; i < dirsToGoUp.size(); i++)
            goDirUp();

        // make absolute path given to this function relative
        // erase path, we are already in
        vDir.erase(0, currDir.length());
        // and add relativ marker
        if (vDir[0] == '/')
            vDir.erase(0, 1);
        //vDir = "./" + vDir;

        // changing to to dir now
        if (!changeDir(vDir))
        {
            // cahanging faild, so reenter folder we were in
            // (and hope it does still exist)
            for (unsigned int i = 0; i < dirsToGoUp.size(); i++)
                goDirDown(dirsToGoUp.at(i));

            return false;
        }
        else 
            return true;
    }
    // path is relativ
    else if (vDir.substr(0, 2) == "./")
    {
        // strip relative sing
        vDir.erase(0, 2);
        return this->changeDir(vDir);
    }
    // step a dir up and then change dir
    else if (vDir.substr(0, 3) == "../")
    {
        string currDir = this->currVirtualDir;      // tmp copy of current dir

        // remove trailing slash from current dir
        int delimiterPos;
        if ((delimiterPos = currDir.find_last_of('/')) == currDir.length() -1)
        {   
            currDir.erase(delimiterPos);
            delimiterPos = currDir.find_last_of('/');
        }

        // any more folders to go up at all?
        if (currDir.length() == 0)
            return false;
        
        // remember folder, we are in now
        string savedDir = currDir.substr(delimiterPos +1);
                
        // remove first '../' from given path
        vDir.erase(0, 3);

        // step dir up and try to change to remaining path
        if (this->goDirUp() &&
            this->changeDir(vDir))
        {
            return true;
        }
        else
        {
            // changing failed, so reenter previously left folder
            this->goDirDown(savedDir);
            return false;
        }
    }
    // elsewise assume that path is meant relative elsewise
    
    // vDir should be realtiv path here

    // try to change to the wanted dir 
    bool successfull = true;        // did an error occur?
    string savedPath = this->currVirtualDir;

    // folder by folder
    int pos = 0;
    while ((pos = vDir.find('/')) != string::npos)
    {
        // seperate the first folder
        string tmp = vDir.substr(0, pos +1);
        vDir.erase(0, pos +1);

        // and try to enter
        if (!goDirDown(tmp))
        {
            successfull = false;
            break;
        }
    }

    // failed to enter one of the folders
    // in case of an error, leave all entered folders again
    if (!successfull)
    {
        this->changeDir(savedPath);
        return false;
    }
    else
        return true;
}


//
// append final backslash to path if non existant
//
void directoryObj::finalBackSlashToPath()
{
    if (this->currDir.length() == 0)
        this->currDir = '\\';
    else if (this->currDir.at(this->currDir.size() -1) != '\\')
        this->currDir += '\\';
}

//
// append final slash to path if non existant
//
void directoryObj::finalSlashToPath()
{
    finalSlashToPath(this->currVirtualDir);
}
void directoryObj::finalSlashToPath(string &path)
{
    if (path.length() == 0)
        path = '/';
    else if (path.at(path.size() -1) != '/')
        path += '/';
}


//
// converts a given filename in current folder to its full path
//
bool directoryObj::fileNameToFullPath(string &fileName)
{
    // a filename shouldn't contain any more dir seperators :)
    if (fileName.find('/') != string::npos ||
        fileName.find('\\') != string::npos)
        return false;

    // ensure curr path ends whith backslash and append path to it. done :)
    finalBackSlashToPath();
    fileName = this->currDir + fileName;
    return true;
}


//
// converts a virtual to a real path
//
bool directoryObj::getRealPath(string &virtualPath)
{
    string path, fileName;  // split path
    int pos;                // position of last path delimiter

    // firefox (Build identifier: Mozilla/5.0 (Windows; U; Windows NT 6.0; de; rv:1.9.0.8) Gecko/2009032609 Firefox/3.0.8 (.NET CLR 3.5.30729))
    // ends filename with slash.. weird...
    if ((pos = virtualPath.find_last_of('/')) == virtualPath.length() -1)
    {
        virtualPath.erase(pos);
        pos = virtualPath.find_last_of('/');
    }

    // is given path a filename or a more complex path?
    if ((pos = virtualPath.find_last_of('/')) == string::npos)
    {
        // just a filename
        path = "";
        fileName = virtualPath;
    }
    else
    {
        // split given path
        path = virtualPath.substr(0, pos);
        fileName = virtualPath.substr(pos +1);
    }

    // remember folder, we are in
    string savedVirtualDir = this->currVirtualDir;
    // try to change to given folder
    bool retVal = this->changeDir(path);

    // ensure real path contains trailing backslash
    this->finalBackSlashToPath();
    // assemble requested real path
    virtualPath = this->currDir + fileName;

    // and restore our current folder
    this->changeDir(savedVirtualDir);

    return retVal;
}
